import Foundation
import Vapor

class MigrationController: RouteCollection {
    func boot(routes: RoutesBuilder) throws {
        let migrationRoutes = routes.grouped("api", "migration")

        let tokenAuthMiddleware = Token.authenticator()
        let guardAuthMiddleware = User.guardMiddleware()
        let tokenAuthGroup = migrationRoutes.grouped(tokenAuthMiddleware, guardAuthMiddleware)

        tokenAuthGroup.post(use: migrationHandler)
        tokenAuthGroup.post("update-locations", use: updateLocationsHandler)
    }

    func migrationHandler(_ req: Request) throws -> EventLoopFuture<[Restaurant]> {
        let data = try req.content.decode([MigrationData].self)


        let getOrCreateCatgegorySpeciality = TagCategory.getOrCreateTagCategory(name: "speciality", req)
        let getOrCreateTagSpeciality = getOrCreateCatgegorySpeciality.flatMap { tagCategory in
            Array(Set(data.compactMap { $0.speciality })).map { Tag.getOrCreateTag(name: $0, categoryID: tagCategory.id ?? UUID(), req) }.flatten(on: req.eventLoop)
        }

        let getOrCreateCatgegoryNeighbourhood = TagCategory.getOrCreateTagCategory(name: "neighbourhood", req)
        let getOrCreateTagNeighbourhood = getOrCreateCatgegoryNeighbourhood.flatMap { tagCategory in
            Array(Set(data.compactMap { $0.neighbourhood })).map { Tag.getOrCreateTag(name: $0, categoryID: tagCategory.id ?? UUID(), req) }.flatten(on: req.eventLoop)
        }

        let getOrCreateCatgegoryType = TagCategory.getOrCreateTagCategory(name: "type", req)
        let getOrCreateTagType = getOrCreateCatgegoryType.flatMap { tagCategory in
            Array(Set(data.compactMap { $0.type })).map { Tag.getOrCreateTag(name: $0, categoryID: tagCategory.id ?? UUID(), req) }.flatten(on: req.eventLoop)
        }

        let allTags = [getOrCreateTagSpeciality, getOrCreateTagNeighbourhood, getOrCreateTagType].flatten(on: req.eventLoop).map { arrayOfArrayTags in
            arrayOfArrayTags.flatMap { $0 }
        }


        let createRestaurants = allTags.flatMap { tags in
            data.map { restaurantToMigrate -> EventLoopFuture<Restaurant> in
                let restaurant = Restaurant(name: restaurantToMigrate.name, info: restaurantToMigrate.info, priceRange: restaurantToMigrate.priceRange, address: restaurantToMigrate.address, phone: restaurantToMigrate.phone)
                let tagsToAdd = tags.filter { $0.name == restaurantToMigrate.speciality || $0.name == restaurantToMigrate.neighbourhood || $0.name == restaurantToMigrate.type }
                return restaurant.save(on: req.db)
                    .flatMap { _ in
                    restaurant.$tags.attach(tagsToAdd, on: req.db)
                    }.map { restaurant }
            }.flatten(on: req.eventLoop)
        }

        return createRestaurants
    }

    func updateLocationsHandler(_ req: Request) throws -> EventLoopFuture<[Restaurant]> {
        let restaurantsQuery = Restaurant.query(on: req.db).all()
        let googlePlacesRequests = restaurantsQuery.flatMap { restaurants -> EventLoopFuture<[Restaurant]> in
            restaurants.compactMap { restaurant -> EventLoopFuture<Restaurant>? in
                let input = "\(restaurant.name),\(restaurant.address ?? "")"
                    .replacingOccurrences(of: " ", with: "%20")
                    .replacingOccurrences(of: "|", with: "%20")
                    .replacingOccurrences(of: "'", with: "%20")
                    .replacingOccurrences(of: "’", with: "%20")
                    .folding(options: .diacriticInsensitive, locale: .current)
                    .folding(options: .caseInsensitive, locale: .current)
                return try? GooglePlacesAPIManager.shared.findPlace(input: input, req).flatMap { response in
                    restaurant.lat = response.candidates.first?.geometry?.location?.lat
                    restaurant.lng = response.candidates.first?.geometry?.location?.lng
                    restaurant.googlePlacesID = response.candidates.first?.place_id
                    return restaurant.save(on: req.db).map { restaurant }
                }
            }.flatten(on: req.eventLoop)
        }

        return googlePlacesRequests
    }

}

struct MigrationData: Codable {
    let name: String
    let info: String?
    let priceRange: String?
    let address: String?
    let phone: String?
    let type: String?
    let speciality: String?
    let veganVegetarian: String?
    let neighbourhood: String?

    enum CodingKeys: String, CodingKey {
        case name = "Nombre"
        case info = "Información"
        case priceRange = "Precio medio"
        case address = "Dirección"
        case phone = "Teléfono"
        case type = "Tipo de comida"
        case speciality = "Especialidad"
        case veganVegetarian = "Vegano Vegetariano"
        case neighbourhood = "Barrio"
    }
}
