import Foundation
import Vapor
import Fluent


class RestaurantsController: RouteCollection {
    func boot(routes: RoutesBuilder) throws {
        let restaurantsRoutes = routes.grouped("api", "restaurants")
        restaurantsRoutes.get(":restaurantID", use: getHandler)
        restaurantsRoutes.get(":restaurantID","tags", use: getTagsHandler)

        let tokenAuthMiddleware = Token.authenticator()
        let guardAuthMiddleware = User.guardMiddleware()
        let tokenAuthGroup = restaurantsRoutes.grouped(tokenAuthMiddleware, guardAuthMiddleware)

        tokenAuthGroup.get(use: getAllHandler)
        tokenAuthGroup.post(use: createHandler)
        tokenAuthGroup.delete(":restaurantID", use: deleteHandler)
        tokenAuthGroup.put(":restaurantID", use: updateHandler)
        tokenAuthGroup.get("favorites", use: getFavoriteRestaurantsHandler)
        tokenAuthGroup.post("favorites",":restaurantID", use: postFavoriteRestaurantHandler)
        tokenAuthGroup.delete("favorites",":restaurantID", use: removeFavoriteRestaurantHandler)

    }

    func getAllHandler(_ req: Request) async throws -> [RestaurantResponse] {
        let restaurants = try await Restaurant.query(on: req.db).all()
        let response = try await restaurants.asyncMap({ restaurant -> RestaurantResponse in
            let tags = try await restaurant.$tags.get(on: req.db)
            try await tags.asyncMap { tag in
                try await tag.$tagCategory.get(on: req.db)
            }
            let restaurantResponse = try await restaurant.transformToResponse(req)
            return restaurantResponse
        })
        return response
    }

    func getHandler(_ req: Request) throws -> EventLoopFuture<Restaurant> {
        Restaurant.find(req.parameters.get("restaurantID"), on: req.db).unwrap(or: Abort(.notFound))
    }

    func createHandler(_ req: Request) throws -> EventLoopFuture<Restaurant> {
        let data = try req.content.decode(CreateRestaurantData.self)
        let tagsQuery = Tag.query(on: req.db).filter(\.$id ~~ (data.tagsID ?? [])).all()
        let restaurant = Restaurant(name: data.name, info: data.info, priceRange: data.priceRange, address: data.address, phone: data.phone)
        return restaurant.save(on: req.db).and(tagsQuery)
            .flatMap { (_, tags) in
                restaurant.$tags.attach(tags, on: req.db)
            }.map { restaurant }
    }

    func deleteHandler(_ req: Request) throws -> EventLoopFuture<HTTPStatus> {
        Restaurant.find(req.parameters.get("restaurantID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { restaurant in
                restaurant.delete(on: req.db).transform(to: .noContent)
            }
    }

    func updateHandler(_ req: Request) throws -> EventLoopFuture<Restaurant> {
        let data = try req.content.decode(CreateRestaurantData.self)
        let updatedRestaurant = try req.content.decode(Restaurant.self)
        let tagsQuery = Tag.query(on: req.db).filter(\.$id ~~ (data.tagsID ?? [])).all()
        return Restaurant.find(req.parameters.get("restaurantID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .and(tagsQuery)
            .flatMap { (restaurant, tags) in
                restaurant.name = updatedRestaurant.name
                restaurant.info = updatedRestaurant.info
                restaurant.priceRange = updatedRestaurant.priceRange
                restaurant.address = updatedRestaurant.address
                restaurant.phone = updatedRestaurant.phone
                return restaurant.$tags.attach(tags, on: req.db).flatMap { _ in
                    restaurant.save(on: req.db).map { restaurant }
                }
            }
    }

    func getTagsHandler(_ req: Request) throws -> EventLoopFuture<[Tag]> {
        Restaurant.find(req.parameters.get("restaurantID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { restaurant in
                restaurant.$tags.get(on: req.db)
            }
    }

    func getFavoriteRestaurantsHandler(_ req: Request) async throws -> [RestaurantResponse] {
        let user = try req.auth.require(User.self)

        let restaurants = try await user.$favoriteRestaurants.get(on: req.db)
        let response = try await restaurants.asyncMap({ restaurant -> RestaurantResponse in
            let tags = try await restaurant.$tags.get(on: req.db)
            try await tags.asyncMap { tag in
                try await tag.$tagCategory.get(on: req.db)
            }
            let restaurantResponse = try await restaurant.transformToResponse(req)
            return restaurantResponse
        })
        return response
    }

    func postFavoriteRestaurantHandler(_ req: Request) async throws -> RestaurantResponse {
        let user = try req.auth.require(User.self)
        guard let restaurant = try await Restaurant.find(req.parameters.get("restaurantID"), on: req.db) else {
            throw Abort(.notFound)
        }


        let userFavorites = try await user.$favoriteRestaurants.get(on: req.db)

        if userFavorites.contains(where: { $0.id?.uuidString ?? "" == restaurant.id?.uuidString ?? "" }) {
            return try await restaurant.transformToResponse(req)
        }

        try await user.$favoriteRestaurants.attach(restaurant, on: req.db)
        let tags = try await restaurant.$tags.get(on: req.db)
        try await tags.asyncMap { tag in
            try await tag.$tagCategory.get(on: req.db)
        }
        return RestaurantResponse(restaurant: restaurant, isFavorite: true)
    }

    func removeFavoriteRestaurantHandler(_ req: Request) async throws -> RestaurantResponse {
        let user = try req.auth.require(User.self)
        guard let restaurant = try await Restaurant.find(req.parameters.get("restaurantID"), on: req.db) else {
            throw Abort(.notFound)
        }

        let userFavorites = try await user.$favoriteRestaurants.get(on: req.db)

        if !userFavorites.contains(where: { $0.id?.uuidString ?? "" == restaurant.id?.uuidString ?? "" }) {
            return try await restaurant.transformToResponse(req)
        }

        try await user.$favoriteRestaurants.detach([restaurant], on: req.db)
        let tags = try await restaurant.$tags.get(on: req.db)
        try await tags.asyncMap { tag in
            try await tag.$tagCategory.get(on: req.db)
        }
        return RestaurantResponse(restaurant: restaurant, isFavorite: false)
    }

}


struct CreateRestaurantData: Content {
    let name: String
    let info: String?
    let priceRange: String?
    let address: String?
    let phone: String?
    let tagsID: [UUID]?
}
