import Foundation
import Vapor

class TagCategoriesController: RouteCollection {

    func boot(routes: RoutesBuilder) throws {
        let tagCategoriesRoutes = routes.grouped("api", "tag-categories")
        tagCategoriesRoutes.get(use: getAllHandler)
        tagCategoriesRoutes.get(":tagCategoryID", use: getHandler)
        tagCategoriesRoutes.get(":tagCategoryID", "tags", use: getTagHandler)

        let tokenAuthMiddleware = Token.authenticator()
        let guardAuthMiddleware = User.guardMiddleware()
        let tokenAuthGroup = tagCategoriesRoutes.grouped(tokenAuthMiddleware, guardAuthMiddleware)

        tokenAuthGroup.post(use: createHandler)
        tokenAuthGroup.delete(":tagCategoryID", use: deleteHandler)
        tokenAuthGroup.put(":tagCategoryID", use: updateHandler)
    }

    func getAllHandler(_ req: Request) throws -> EventLoopFuture<[TagCategory]> {
        TagCategory.query(on: req.db).all().flatMap { tagCategories in
            tagCategories.map { tagCategory in
                tagCategory.$tags.get(on: req.db)
                    .map { _ in tagCategory }
            }
            .flatten(on: req.eventLoop)
        }
    }

    func getHandler(_ req: Request) throws -> EventLoopFuture<TagCategory> {
        TagCategory.find(req.parameters.get("tagCategoryID"), on: req.db).unwrap(or: Abort(.notFound))
    }

    func createHandler(_ req: Request) throws -> EventLoopFuture<TagCategory> {
        let data = try req.content.decode(CreateTagCategoryData.self)
        let tagCategory = TagCategory(name: data.name)
        return tagCategory.save(on: req.db).map{ tagCategory }
    }

    func deleteHandler(_ req: Request) throws -> EventLoopFuture<HTTPStatus> {
        TagCategory.find(req.parameters.get("tagCategoryID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { tagCategory in
                tagCategory.delete(on: req.db).transform(to: .noContent)
            }
    }

    func updateHandler(_ req: Request) throws -> EventLoopFuture<TagCategory> {
        let updatedTagCategory = try req.content.decode(TagCategory.self)
        return TagCategory.find(req.parameters.get("tagCategoryID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { tagCategory in
                tagCategory.name = updatedTagCategory.name
                return tagCategory.save(on: req.db).map { tagCategory }
            }
    }

    func getTagHandler(_ req: Request) throws -> EventLoopFuture<[Tag]> {
        return TagCategory.find(req.parameters.get("tagCategoryID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { tagCategory in
                tagCategory.$tags.get(on: req.db)
            }

    }
}

struct CreateTagCategoryData: Content {
    let name: String
}
