import Foundation
import Vapor

class TagsController: RouteCollection {

    func boot(routes: RoutesBuilder) throws {
        let tagsRoutes = routes.grouped("api", "tags")
        tagsRoutes.get(use: getAllHandler)
        tagsRoutes.get(":tagID", use: getHandler)
        tagsRoutes.get(":tagID","tagCategory", use: getTagCategoryHandler)
        tagsRoutes.get(":tagID","restaurants", use: getRestaurantHandler)

        let tokenAuthMiddleware = Token.authenticator()
        let guardAuthMiddleware = User.guardMiddleware()
        let tokenAuthGroup = tagsRoutes.grouped(tokenAuthMiddleware, guardAuthMiddleware)

        tokenAuthGroup.post(use: createHandler)
        tokenAuthGroup.delete(":tagID", use: deleteHandler)
        tokenAuthGroup.put(":tagID", use: updateHandler)
        tokenAuthGroup.get("favorite", use: getFavoriteTags)
        tokenAuthGroup.post("favorite", use: postFavoriteTags)

    }

    func getAllHandler(_ req: Request) async throws -> [TagResponse] {
        let tags = try await Tag.query(on: req.db).all()
        let tagsResponse = try await tags.asyncMap { tag -> TagResponse in
            let response = try await tag.transformToResponse(req)
            return response
        }
        return tagsResponse
    }

    func getHandler(_ req: Request) throws -> EventLoopFuture<Tag> {
        Tag.find(req.parameters.get("tagID"), on: req.db).unwrap(or: Abort(.notFound))
    }

    func createHandler(_ req: Request) throws -> EventLoopFuture<Tag> {
        let data = try req.content.decode(CreateTagData.self)
        let tag = Tag(name: data.name, tagCategoryID: data.tagCategoryID)
        return tag.save(on: req.db).map { tag }
    }

    func deleteHandler(_ req: Request) throws -> EventLoopFuture<HTTPStatus> {
        Tag.find(req.parameters.get("tagID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { tag in
                tag.delete(on: req.db).transform(to: .noContent)
            }
    }

    func updateHandler(_ req: Request) throws -> EventLoopFuture<Tag> {
        let updatedTag = try req.content.decode(Tag.self)
        return Tag.find(req.parameters.get("tagID"), on: req.db)
            .unwrap(or: Abort(.notFound))
            .flatMap { tag in
                tag.name = updatedTag.name
                return tag.save(on: req.db).map { tag }
            }
    }

    func getTagCategoryHandler(_ req: Request) throws -> EventLoopFuture<TagCategory> {
        Tag.find(req.parameters.get("tagID"), on: req.db).unwrap(or: Abort(.notFound)).flatMap { tag in
            tag.$tagCategory.get(on: req.db)
        }
    }

    func getRestaurantHandler(_ req: Request) throws -> EventLoopFuture<[Restaurant]> {
        Tag.find(req.parameters.get("tagID"), on: req.db).unwrap(or: Abort(.notFound)).flatMap { tag in
            tag.$restaurants.get(on: req.db)
        }
    }

    func postFavoriteTags(_ req: Request) throws -> EventLoopFuture<[Tag]> {
        let user = try req.auth.require(User.self)
        let request = try req.content.decode(PostFavoriteRequest.self)

        let findedTagsFuture = request.tags.map { Tag.find($0, on: req.db).unwrap(or: Abort(.notFound)) }.flatten(on: req.eventLoop)

        let userFavoriteTags = FavoriteTags.query(on: req.db).all().flatMap { tags  in
            tags.map { $0.$user.get(on: req.db) }.flatten(on: req.eventLoop).map { _ -> [FavoriteTags] in
                tags.filter { $0.user.id == user.id }
            }
        }

        let deletePreviousUserFavoriteTags = userFavoriteTags.flatMap { tags in
            tags.map { $0.delete(on: req.db) }.flatten(on: req.eventLoop)
        }

        let createUserFavoriteTags = findedTagsFuture.flatMap { tags in
            user.$favoriteTags.attach(tags, on: req.db).map { tags }
        }

        return deletePreviousUserFavoriteTags.and(createUserFavoriteTags).map { $1 }
    }


    func getFavoriteTags(_ req: Request) throws -> EventLoopFuture<[Tag]> {
        let user = try req.auth.require(User.self)
        return user.$favoriteTags.get(on: req.db)
    }
}

struct CreateTagData: Content {
    let name: String
    let tagCategoryID: UUID
}


struct PostFavoriteRequest: Content {
    let tags: [UUID]
}
