import Foundation
import FluentKit
import Vapor

final class Restaurant: Model {

    @ID
    var id: UUID?

    @Field(key: "name")
    var name: String

    @Field(key: "info")
    var info: String?

    @Field(key: "priceRange")
    var priceRange: String?

    @Field(key: "address")
    var address: String?

    @Field(key: "phone")
    var phone: String?

    @Field(key: "lng")
    var lng: Double?

    @Field(key: "lat")
    var lat: Double?

    @Field(key: "googlePlacesID")
    var googlePlacesID: String?

    @Siblings(through: RestaurantTagPivot.self, from: \.$restaurant, to: \.$tag)
    var tags: [Tag]

    @Siblings(through: Visit.self, from: \.$restaurant, to: \.$user)
    var visitors: [User]

    init() { }

    init(id: UUID? = nil, name: String, info: String? = nil, priceRange: String? = nil, address: String? = nil, phone: String? = nil, lat: Double? = nil, lng: Double? = nil, googlePlacesID: String? = nil) {
        self.id = id
        self.name = name
        self.info = info
        self.priceRange = priceRange
        self.address = address
        self.phone = phone
        self.lat = lat
        self.lng = lng
        self.googlePlacesID = googlePlacesID
    }

    static let schema: String = "restaurants"

    

}

extension Restaurant: Content {
    func transformToResponse(_ req: Request) async throws -> RestaurantResponse {
        let _ = try await self.$tags.get(on: req.db)
        let user = try req.auth.require(User.self)

        let favoriteRestaurants = try await user.$favoriteRestaurants.get(on: req.db)

        return RestaurantResponse(restaurant: self, isFavorite: favoriteRestaurants.contains { self.id?.uuidString ?? "" == $0.id?.uuidString ?? "" })
    }
}

extension Restaurant {
    static func requestRestaurantsWithTags(_ req: Request) -> EventLoopFuture<[(Restaurant,[Tag])]> {
        return Restaurant.query(on: req.db).all()
            .flatMap { restaurants in
                restaurants.map { restaurant in
                    restaurant.$tags.get(on: req.db).map { (restaurant,$0)}
                }.flatten(on: req.eventLoop).map { $0.map { ($0.0, $0.1.sorted { $0.id?.uuidString ?? "" > $1.id?.uuidString ?? "" } ) } }
        }

    }
}

struct RestaurantResponse: Content {
    let id: UUID?
    let name: String
    let info: String?
    let priceRange: String?
    let address: String?
    let phone: String?
    let tags: [TagResponse]
    let lat: Double?
    let lng: Double?
    let isFavorite: Bool?

    init(restaurant: Restaurant, isFavorite: Bool? = nil) {
        self.id = restaurant.id
        self.name = restaurant.name
        self.info = restaurant.info
        self.priceRange = restaurant.priceRange
        self.address = restaurant.address
        self.phone = restaurant.phone
        self.tags = restaurant.tags.map {TagResponse(tag: $0)}
        self.lat = restaurant.lat
        self.lng = restaurant.lng
        self.isFavorite = isFavorite
    }
}
