import Foundation
import Vapor
import Fluent

final class User: Model {

    @ID
    var id: UUID?

    @Field(key: "name")
    var name: String

    @Field(key: "surname")
    var surname: String

    @Field(key: "email")
    var email: String

    @Field(key: "password")
    var password: String

    @Siblings(through: Visit.self, from: \.$user, to: \.$restaurant)
    var visitedRestaurants: [Restaurant]

    @Siblings(through: FavoriteTags.self, from: \.$user, to: \.$tag)
    var favoriteTags: [Tag]

    @Siblings(through: FavoriteRestaurants.self, from: \.$user, to: \.$restaurant)
    var favoriteRestaurants: [Restaurant]

    init() { }

    init(id: UUID? = nil, name: String, surname: String, email: String) {
        self.id = id
        self.name = name
        self.surname = surname
        self.email = email
    }

    static let schema: String = "users"

    struct Public: Content {
        var id: UUID?
        var name: String
        var surname: String
        var email: String
    }
}

extension User: Content { }

extension User {

    func convertToPublic() -> User.Public {
        User.Public(id: self.id, name: self.name, surname: self.surname, email: self.email)
    }

    func requestVisitedRestaurantsTagsCount(_ req: Request) -> EventLoopFuture<[(Tag, Int)]> {
        return self.$visitedRestaurants.get(on: req.db).flatMap { (restaurants: Array<Restaurant>) in
            restaurants.map { restaurant in
                restaurant.$tags.get(on: req.db)
            }.flatten(on: req.eventLoop).map { arrayOfArrayOfTags in
                let tags = arrayOfArrayOfTags.flatMap { $0 }
                return Dictionary(grouping: tags) { $0.id }.map { (key, value) in
                    (value.first!, value.count)
                }.sorted { $0.0.id?.uuidString ?? "" > $1.0.id?.uuidString ?? ""}
            }
        }
    }
}

extension EventLoopFuture where Value: User {
    func convertToPublic() -> EventLoopFuture<User.Public> {
        self.map { user in
            user.convertToPublic()
        }
    }
}

extension Collection where Element: User {
    func convertToPublic() -> [User.Public] {
        self.map { $0.convertToPublic() }
    }
}

extension EventLoopFuture where Value == Array<User> {
    func convertToPublic() -> EventLoopFuture<[User.Public]> {
        self.map { $0.convertToPublic() }
    }
}

extension User: ModelAuthenticatable {
    static let usernameKey = \User.$email
    static let passwordHashKey = \User.$password

    func verify(password: String) throws -> Bool {
        try Bcrypt.verify(password, created: self.password)
    }
}

