//
//  Recommender.swift
//  
//
//  Created by Joan Martin on 8/11/22.
//

import Foundation
import Vapor

struct Recommender {
    let restaurants: [Restaurant]
    let req: Request
    let tags: [Tag]

    var user: User? {
        return try? req.auth.require(User.self)
    }

    func getRecommendations() async throws -> [Restaurant] {
        let userFeatureVector = try await getUserFeatureVector()
        let restaurantsFeatureVectors = try await getRestaurantsFeatureVectors()
        return calculateSimilarityMatrix(userFeatureVector: userFeatureVector, restaurantFeatureVectors: restaurantsFeatureVectors).sorted { $0.similarity > $1.similarity }.map { $0.restaurant }
    }

    func getRestaurantsFeatureVectors() async throws -> [RestaurantFeatureVector]  {
        var restaurantsFeatureVectors: [RestaurantFeatureVector] = []

        for restaurant in restaurants {
            let tags = try await restaurant.$tags.get(on: req.db)
            var featureVector = tags.map { FeatureVectorComponent(tag: $0, count: 1) }
            self.tags.forEach { tag in
                if let id = tag.id, !featureVector.contains(where: { $0.tag.id == id }) {
                    featureVector.append(FeatureVectorComponent(tag: tag, count: 0))
                }
            }
            featureVector.sort { $0.tag.id?.uuidString ?? "" > $1.tag.id?.uuidString ?? "" }
            restaurantsFeatureVectors.append(RestaurantFeatureVector(restaurant: restaurant, featureVector: normalize(featureVector: featureVector)))
        }
        return restaurantsFeatureVectors
    }

    func getUserFeatureVector() async throws -> FeatureVector {
        guard let user = user else {
            return []
        }
        // Initialize user feature vector to 0 and sort tags
        var userFeatureVector: FeatureVector = tags.map { FeatureVectorComponent(tag: $0, count: 0) }

        userFeatureVector.sort { $0.tag.id?.uuidString ?? "" > $1.tag.id?.uuidString ?? ""}

        // Get visited restaurants and extract number of visits of user
        let visitedRestaurants = try await user.$visitedRestaurants.get(on: req.db)
        let userTags = try await visitedRestaurants.map { $0.$tags.get(on: req.db) }.flatten(on: req.eventLoop).get().flatMap { $0 }
        let groupedTags = Dictionary(grouping: userTags) { $0.id }
        let userTagsCount = groupedTags.compactMap { (id, tags) -> FeatureVectorComponent? in
            guard let tag = tags.first else { return nil }
            return FeatureVectorComponent(tag: tag, count: Double(tags.count))
        }

        userFeatureVector = userFeatureVector.map({ component in
            if let userTagCount = userTagsCount.first(where: { $0.tag.id?.uuidString ?? "" == component.tag.id?.uuidString ?? "" }) {
                return FeatureVectorComponent(tag: component.tag, count: component.count + userTagCount.count)
            } else {
                return component
            }
        })

        // Get user favorite tags
        let favoriteTags = try await user.$favoriteTags.get(on: req.db)

        userFeatureVector = userFeatureVector.map({ component in
            if let _ = favoriteTags.first(where: { $0.id?.uuidString ?? "" == component.tag.id?.uuidString ?? "" }) {
                return FeatureVectorComponent(tag: component.tag, count: component.count + 10)
            } else {
                return component
            }
        })

        // Get user favorite restaurants
        let favoriteRestaurants = try await user.$favoriteRestaurants.get(on: req.db)
        let favoriteRestaurantsTags = try await favoriteRestaurants
            .asyncMap({ restaurant -> [Tag] in
            return try await restaurant.$tags.get(on: req.db)
        })
            .flatMap { $0 }
        let groupedFavoriteRestaurantsTags = Dictionary(grouping: favoriteRestaurantsTags) { $0.id }
        let userFavoriteRestaurantsTagsCount = groupedFavoriteRestaurantsTags.compactMap { (id, tags) -> FeatureVectorComponent? in
            guard let tag = tags.first else { return nil }
            return FeatureVectorComponent(tag: tag, count: Double(tags.count))
        }

        userFeatureVector = userFeatureVector.map({ component in
            if let favoriteRestaurantTagCount = userFavoriteRestaurantsTagsCount.first(where: { $0.tag.id?.uuidString ?? "" == component.tag.id?.uuidString ?? "" }) {
                return FeatureVectorComponent(tag: component.tag, count: component.count + (favoriteRestaurantTagCount.count * 2))
            } else {
                return component
            }
        })

        return normalize(featureVector: userFeatureVector)
    }

    private func normalize(featureVector: FeatureVector) -> FeatureVector {
        let sum = featureVector.reduce(0.0) { $0 + $1.count }
        if sum == 0.0 {
            return featureVector.map { FeatureVectorComponent(tag: $0.tag, count: 0.0) }
        }
        let normalizedFeatureVector = featureVector.map { FeatureVectorComponent(tag: $0.tag, count: $0.count / sum) }
        return normalizedFeatureVector
    }

    private func dotProduct(first: [Double], second: [Double]) -> Double  {
        guard first.count == second.count else {
            fatalError("Vector space is not the same")
        }

        var result: Double = 0
        for i in 0..<first.count {
            result += first[i] * second[i]
        }
        return sqrt(result)
    }

    private func calculateSimilarityMatrix(userFeatureVector: FeatureVector, restaurantFeatureVectors: [RestaurantFeatureVector]) -> [RestaurantSimilarity] {
        return restaurantFeatureVectors.map { restaurantFeatureVector in
            let simplifiedRestaurantFeatureVector = restaurantFeatureVector.featureVector.map { $0.count }
            let simplifiedUserFeatureVector = userFeatureVector.map { $0.count }
            print(simplifiedRestaurantFeatureVector)
            print(simplifiedUserFeatureVector)

            return RestaurantSimilarity(restaurant: restaurantFeatureVector.restaurant, similarity: dotProduct(first: simplifiedRestaurantFeatureVector, second: simplifiedUserFeatureVector))
        }
    }

    struct FeatureVectorComponent {
        let tag: Tag
        var count: Double
    }

    typealias FeatureVector = [FeatureVectorComponent]

    struct RestaurantFeatureVector {
        let restaurant: Restaurant
        let featureVector: FeatureVector
    }

    struct RestaurantSimilarity {
        let restaurant: Restaurant
        let similarity: Double
    }
}
