//
//  APIService.swift
//  Eating Barna
//
//  Created by Joan Martin on 24/5/22.
//

import Foundation
import Combine

protocol APIService {
    func request<T: Decodable>(with endpoint: Endpoint) async throws -> T
}

protocol Endpoint {

    var scheme: String { get }

    var baseURL: String { get }

    var path: String { get }

    var parameters: [URLQueryItem] { get }

    var body: [String: Any]? { get }

    var headers: [String:String] { get }

    var method: String { get }

    var authoritzationType: AuthoritzationType? { get }

}

enum APIError: Error {
    case decodingError
    case httpError(Int)
    case unknown
}

enum EatingBarnaEndpoint: Endpoint {
    case login(email: String, password: String)
    case getRestaurants
    case getRecommendations
    case postVisit(restaurantID: String)
    case getTagCategories
    case postUser(name: String, surname: String, email: String, password: String)
    case getFavoriteTags
    case postFavoriteTags(tags: [String])
    case postFavoriteRestaurant(id: String)
    case deleteFavoriteRestaurant(id: String)

    private var userDefaultsManager: UserDefaultsManager  {
        return UserDefaultsManager.shared
    }

    var scheme: String {
        switch self {
        default: return "http"
        }
    }

    var baseURL: String {
        switch self {
        default: return "127.0.0.1:8080"
        }
    }

    var path: String {
        switch self {
        case .login:
            return "/api/users/login"
        case .getRestaurants:
            return "/api/restaurants"
        case .getRecommendations:
            return "/api/recommendations"
        case .postVisit(let restaurantID):
            return "/api/visits/\(restaurantID)"
        case .getTagCategories:
            return "/api/tag-categories"
        case .postUser:
            return "/api/users"
        case .getFavoriteTags, .postFavoriteTags:
            return "/api/tags/favorite"
        case .postFavoriteRestaurant(id: let id):
            return "/api/restaurants/favorites/\(id)"
        case .deleteFavoriteRestaurant(id: let id):
            return "/api/restaurants/favorites/\(id)"

        }
    }

    var parameters: [URLQueryItem] {
        switch self {
        default: return []
        }
    }

    var body: [String: Any]? {
        switch self {
        case .postUser(let name, let surname, let email, let password):
            return ["name": name,
                    "surname": surname,
                    "email": email,
                    "password": password]
        case .postFavoriteTags(let tags):
            return ["tags": tags]
        default:
            return nil
        }
    }

    var headers: [String: String] {
        var headers = [String:String]()
        switch self {
        case .postUser, .postFavoriteTags:
            headers["Content-Type"] = "application/json"
        default:
            break;
        }

        if let auth = self.authoritzationType {
            switch auth {
            case .basic(let email, let password):
                guard let loginString = "\(email):\(password)"
                    .data(using: .utf8)?
                    .base64EncodedString() else { return headers }
                headers["Authorization"] = "Basic \(loginString)"
            case .bearer(let token):
                headers["Authorization"] = "Bearer \(token)"
            }
        }
        headers["Accept"] = "*/*"
        headers["Connection"] = "keep-alive"
        headers["Accept-Encoding"] = "gzip, deflate"
        headers["Cache-Control"] = "no-cache"
        return headers
    }

    var method: String {
        switch self {
        case .login, .postVisit, .postUser, .postFavoriteTags, .postFavoriteRestaurant:
            return "POST"
        case .getRestaurants, .getRecommendations, .getTagCategories, .getFavoriteTags:
            return "GET"
        case .deleteFavoriteRestaurant:
            return "DELETE"
        }
    }

    var authoritzationType: AuthoritzationType? {
        switch self {
        case .login(let email, let password):
            return .basic(email: email, password: password)
        case .getRecommendations, .postVisit, .getFavoriteTags, .postFavoriteTags, .getRestaurants, .postFavoriteRestaurant, .deleteFavoriteRestaurant:
            return .bearer(token: userDefaultsManager.getToken() ?? "")
        case .getTagCategories, .postUser:
            return nil
        }
    }
}

enum AuthoritzationType {
    case bearer(token: String)
    case basic(email: String, password: String)
}
